#  Action Script v. 2018-12-12
#
#  Generic Action (grasp,release,mount,unmount,traceon,traceoff, sweptvolumeon, sweptvolumeoff) execution
#  by triggering a specific robot output with a specific value
#
#  Changes
#  -Trace changed to use vcCompactLineSet instead of vcLineSet
#
from vcScript import *
import vcVector
import vcMatrix
global grasp_association
comp = getComponent() 
app = getApplication()
normal_vec = vcVector.new(0,0,1)


def confirmProperty(name,defaultvalue,type=VC_REAL,constraints=VC_PROPERTY_DEFAULT):
     global comp
     prop = comp.getProperty(name)
     if prop == None:
         prop = comp.createProperty(type,name,constraints)
         prop.Value = defaultvalue
     return prop


Verbose = confirmProperty("SignalActions::DisplayMessages",False,VC_BOOLEAN)
ReleaseToWorld = confirmProperty("SignalActions::ReleaseToWorld",False,VC_BOOLEAN)
ReleaseToPhysics = confirmProperty("SignalActions::ReleaseToPhysics",'Never',VC_STRING, VC_PROPERTY_STEP)
ReleaseToPhysics.StepValues = ['Never', 'If Target not Detected', 'Always']
MultiGrasp = confirmProperty("SignalActions::MultiGrasp",False,VC_BOOLEAN)
TraceWidth = confirmProperty("SignalActions::TraceWidth",1,VC_INTEGER)
TraceWidth.IsVisible = True
TraceZOffset = confirmProperty("SignalActions::TraceZOffset",0.,VC_REAL)
group_index = 0
Verbose.Group = group_index = group_index+10
ReleaseToWorld.Group = group_index = group_index+10
ReleaseToPhysics.Group = group_index = group_index+10
MultiGrasp.Group = group_index = group_index+10
TraceWidth.Group = group_index = group_index+10
TraceZOffset.Group = group_index = group_index+10
activeSweptVolumeItems = []
allSweptVolumeItems = []
red_material = app.findMaterial('red')
CONFIGURE_DIALOG_CMD = "showGraspRelease"
ASSOCIATE_GRASPING_TO_OUTPUT = True
ACTIONS_NAME1 = "Actions"
MOUNTPLATE_NAME = "mountplate"
GRASP_CONTAINER = "GraspContainer"
ACTION_BUTTON1 = "SignalActions::Configure"
TRACE_LAYER = "RobotTrace"
TRACE_COLORS = [ "red", "blue", "green", "white", "black", "grey", "yellow", "brown", "cyan", "purple", "ingido", "orange", "pink", "beige", "cream", "khaki" ]
simu = None


class Trace:
  def __init__(self):
    self.Tool = None
    self.ToolName = None
    self.Base = None
    self.BaseName = None
    self.Node = None
    self.Material = None
    self.LineSet = None
    self.Line = None
    self.Port = -1
    self.Robot = None
    self.Light = None
    self.Particles = None
    
        
# Create new action with its properties
def AddAction(actor, name, props):
  action = actor.createAction(name)
  for prop in props:
    if len(prop) > 3:
      p = action.createProperty(prop[0],prop[1],prop[3])
      p.StepValues = prop[2]
      p.Value = prop[2][prop[4]]
    else:
      p = action.createProperty(prop[0],prop[1])
      p.Value = prop[2]
    p.IsVisible = True


def ToolName(index):
  return "TOOL_DATA[%i]" % index


def BaseName(index):
  return "BASE_DATA[%i]" % index


def Tool(toolname):
  for tool in controller.Tools:
    if tool.Name == toolname:
      return tool
  return None


def Base(basename):
  for base in controller.Bases:
    if base.Name == basename:
      return base
  return None


# Configure component if required
def AutoConfigure():
  global comp, controller
  FORCE_RESET_ACTION = False
  exec_list = comp.findBehavioursByType(VC_ROBOTEXECUTOR)
  executor = None
  outputmap = None
  controller = None
  if len(exec_list):
    executor = exec_list[0]
  if executor:
    outputmap = executor.DigitalOutputSignals
    controller = executor.Controller
  if outputmap:
    if outputmap.PortCount < 49:
      outputmap.PortCount = 49
  # find default flange interface
  iface_name = ""
  ifaces = comp.findBehavioursByType(VC_ONETOONEINTERFACE)
  if len(ifaces):
    iface_name = ifaces[len(ifaces)-1].Name
  # Add action container and all default actions if not available yet
  actor = comp.findBehaviour(ACTIONS_NAME1)
  if not actor:
    actor = comp.createBehaviour(VC_ACTIONCONTAINER,ACTIONS_NAME1)
    FORCE_RESET_ACTION = True
  if FORCE_RESET_ACTION:
    actor.clear()
    # Add default actions
    toolnames = []
    basenames = []
    if controller:
      for tool in controller.Tools:
        toolnames.append(tool.Name)
      if len(toolnames) < 16:
        for i in range(len(toolnames),17):
          toolnames.append("*NULL*")
      for base in controller.Bases:
         basenames.append(base.Name)
      if len(basenames) < 32:
        for i in range(len(toolnames),33):
          basenames.append("*NULL*")  
    else:
      for i in range(1,17):
        toolnames.append(ToolName(i))
      for i in range(1,33):
        basenames.append(BaseName(i))
    for i in range(1,17):
      if toolnames[i-1] != "*NULL*":
        AddAction(actor, "Grasp", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", True), (VC_STRING,"UsingTool",toolnames[i-1]), (VC_VECTOR,"DetectionVolumeSize",vcVector.new(50,50,50) )])
    for i in range(1,17):
      if toolnames[i-1] != "*NULL*":
        AddAction(actor, "Release", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", False), (VC_REAL,"GravityDirection",25.0)])
    for i in range(17,33):
      if toolnames[i-17] != "*NULL*":
        AddAction(actor, "TraceOn", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", True), (VC_STRING,"UsingTool",toolnames[i-17]), (VC_STRING,"Attached","%s::%s" % (comp.Name, comp.Name)), (VC_STRING,"Color",TRACE_COLORS[(i-1) % 16])])
    for i in range(17,33):
      if toolnames[i-17] != "*NULL*":
        AddAction(actor, "TraceOff", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", False), (VC_STRING,"UsingTool",toolnames[i-17])])
    for i in range(33,49):
      if toolnames[i-33] != "*NULL*":
        AddAction(actor, "MountTool", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", True), (VC_STRING,"UsingTool",toolnames[i-33]), (VC_VECTOR,"DetectionVolumeSize",vcVector.new(50,50,50) ),(VC_STRING,"Interface",iface_name)])
    for i in range(33,49):
      if toolnames[i-33] != "*NULL*":
        AddAction(actor, "UnmountTool", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", False), (VC_REAL,"GravityDirection",25.0),(VC_STRING,"Interface",iface_name)])
    for i in range(49,81):
      if basenames[i-49] != "*NULL*":
        AddAction(actor, "TraceOn", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", True), (VC_STRING,"UsingTool",basenames[i-49]), (VC_STRING,"Attached","%s::%s" % (comp.Name, controller.FlangeNode.Name)), (VC_STRING,"Color",TRACE_COLORS[(i-1) % 16])])
    for i in range(49,81):
      if basenames[i-49] != "*NULL*":
        AddAction(actor, "TraceOff", [ (VC_INTEGER, "Port", i), (VC_BOOLEAN, "Value", False), (VC_STRING,"UsingTool",basenames[i-49])])
    AddAction(actor, "SweptVolumeOn", [ 
    (VC_INTEGER, "Port", 81), 
    (VC_BOOLEAN, "Value", True),
    (VC_STRING, "SweptMethod", ['HullFromHulls','HullFromSilhouettes','Lofting'], VC_PROPERTY_STEP, 0),
    ('List<Ref<Node>>', "SweptGeometry", [comp]),
    ('Ref<Material>', "Material", red_material),
    (VC_STRING, "AttachTo", ['Node','Closest','NewComponent'], VC_PROPERTY_STEP, 2),
    ('Ref<Node>', "TargetNode", comp),
    (VC_STRING,"UsingTool",toolnames[0]),
    (VC_VECTOR,"DetectionVolumeSize",vcVector.new(50,50,50) )])
    AddAction(actor, "SweptVolumeOff", [ (VC_INTEGER, "Port", 81), (VC_BOOLEAN, "Value", False), (VC_BOOLEAN, "StoreGeometry", False)])
  # Add configuration editor button
  if not comp.getProperty(ACTION_BUTTON1):
    prop = comp.createProperty(VC_BUTTON, ACTION_BUTTON1)
  # Add graspcontainer to flangenode if not created yet
  container = comp.findBehaviour(GRASP_CONTAINER)
  if not container:
    flange = comp.findNode(MOUNTPLATE_NAME)
    if flange:
      container = flange.createBehaviour(VC_COMPONENTCONTAINER,GRASP_CONTAINER)
    else:
      print "Configuration error: no flange node (%s) found in order to create a container for it (%s)" % (MOUNTPLATE_NAME,GRASP_CONTAINER)


# Display Grasp & Release configuration GUI
def GraspReleaseConfiguration(args):
  cmd = app.findCommand(CONFIGURE_DIALOG_CMD)
  cmd.execute("All:Actions")


# Configure actions and their visibility
AutoConfigure()
if app.findCommand(CONFIGURE_DIALOG_CMD):
  prop1 = comp.getProperty(ACTION_BUTTON1)
  prop1.OnChanged = GraspReleaseConfiguration
  prop1.IsVisible = True
else:
  prop1 = comp.getProperty(ACTION_BUTTON1)
  prop1.IsVisible = False


def OutputTriggered( map, port, value ):
  global comp, executor, outputmap, all_actions
  # rint "Output %i set to value %i" % (port,value)
  # Find if there is an action for this signal
  for actions in all_actions:
    action = None
    for act in actions.Actions:
      # rint act.Name, act.Port, act.Value
      if act.Port == port and act.Value == value:
        action = act
        break
    if not action:
      return
    ret = ""
    if action.Name == "Grasp":
      ret = Grasp( port, action.UsingTool, action.DetectionVolumeSize )
    elif action.Name == "Release":
      ret = Release( port, action.GravityDirection )
    elif action.Name == "MountTool":
      ret = MountTool( port, action.UsingTool, action.DetectionVolumeSize, action.Interface )
    elif action.Name == "UnmountTool":
      ret = UnmountTool( port, action.GravityDirection, action.Interface )
    elif action.Name == "TraceOn":
      ret = TraceOn( port, action.UsingTool, action.Attached, action.Color )
    elif action.Name == "TraceOff":
      ret = TraceOff( action.UsingTool )
    elif action.Name == "SweptVolumeOn":
      ret = SweptVolumeOn( port, action )
    elif action.Name == "SweptVolumeOff":
      ret = SweptVolumeOff( port, action.StoreGeometry )
    if Verbose.Value:
      print "%s: Action \'%s\' executed (%s) " % (comp.Name, action.Name,ret)


# Initialize Grasp & release action implementation
def OnStart():
  global simu, comp, robot, executor, outputmap, all_actions, grasp_container, grasp_association, traces, trace_lines, grasped_components, activeSweptVolumeItems, allSweptVolumeItems, controller
  # Initialize signal map entries wit the attached action entry
  grasped_components = []
  grasp_association = {}
  simu = getSimulation()
  for sv in allSweptVolumeItems:
    sv.clear()
  activeSweptVolumeItems = []
  allSweptVolumeItems = []
  traces = []
  trace_lines = []
  grasp_container = comp.findBehaviour(GRASP_CONTAINER)
  exec_list = comp.findBehavioursByType(VC_ROBOTEXECUTOR)
  executor = None
  outputmap = None
  if len(exec_list):
    executor = exec_list[0]
  if not executor:
    return
  robot = executor.Controller
  controller = executor.Controller
  outputmap = executor.DigitalOutputSignals
  if not outputmap:
    return
  outputmap.OnSignalTrigger = OutputTriggered
  all_actions = comp.findBehavioursByType(VC_ACTIONCONTAINER) 


def DetectionHelper( tool_name, bound ):
  global simu, comp, robot, grasp_container
  # Define bound corner points when original bound vector is thought as bound box diagonal vector
  v1 = bound*(-0.5)
  v2 = bound*0.5
  # Following code sets a specific tool and extracts tool matrix in world coordinate system
  # This transformation is then applied to the bound box defining volume.
  simu.update()
  tool = Tool(tool_name)
  if tool:
    tnode = tool.Node
    tmatrix = tool.PositionMatrix
  else:
    tnode = robot.FlangeNode
    tmatrix = vcMatrix.new()
  m = tnode.WorldPositionMatrix * tmatrix
  v = m.P
  # rint "Detection pos (%f, %f, %f)" % (v.X, v.Y, v.Z)
  # a) Create volume bbox detector and test if geometry bbox is fully inside the detection volume
  vol_detector = getSimulation().newVolumeDetector()
  vol_detector.TestMethod = VC_INSIDE
  vol_detector.Corner1 = v1
  vol_detector.Corner2 = v2
  vol_detector.Transformation = m
  vol_detector.NodeList = [ (simu.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE), (comp, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE  ) ]
  hit = vol_detector.testOneCollision()
  if hit:
    return vol_detector.getHitNode(0).Component
  # b) If no geometry bbox was fully inside the detection volume, create new colision detector
  # using volume as detection geometry
  coll_detector = getSimulation().newCollisionDetector()
  coll_detector.NodeListA = [ (simu.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) , (comp, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE )]
  hit = coll_detector.testOneBBoxCollision(v1,v2,m)
  if hit:
    return coll_detector.getHitNodeA(0).Component
  else:
    return None


def Grasp( output, tool_name, bound ):
  global simu, comp, robot, grasp_container, grasp_association, grasped_components
  grasp_cont = grasp_container
  tool = Tool(tool_name)
  if tool:
    tnode = tool.Node
    if tnode:
      container = tnode.getBehaviour( GRASP_CONTAINER )
      if not container:
        contname = '%s_%s_%s' % (GRASP_CONTAINER, tnode.Component.Name, tnode.Name )
        contname = contname.replace(' ', '')
        container = tnode.findBehaviour( contname )
        if not container:
          container = tnode.createBehaviour( VC_COMPONENTCONTAINER, contname )
    if container:
      grasp_cont = container
  if MultiGrasp.Value:
    count = 0
    list = []
    while True:
      part = DetectionHelper( tool_name, bound )
      if not part:
        break
      list.append(part)
      root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
      for root_phys_entity in root_phys_entities:
        physicsType = root_phys_entity.getProperty('Physics Type')
        if physicsType.Value == 0:
          physicsType.Value = 2 # switch to kinematic
      grasp_cont.grab(part)
      grasped_components.append( part )
      count += 1
    grasp_association[output] = list
    if count > 0:
      return "Ok (%i components grasped)" % count
    else:
      return "No component detected to grasp"
  else:
    part = DetectionHelper( tool_name, bound )
    if part:
      grasp_association[output] = part
      root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
      for root_phys_entity in root_phys_entities:
        physicsType = root_phys_entity.getProperty('Physics Type')
        if physicsType.Value == 0:
          physicsType.Value = 2 # switch to kinematic
      grasp_cont.grab(part)
      grasped_components.append( part )
      return "Ok"
    else:
      return "No component detected to grasp"


def MountTool( output, tool_name, bound, iface_name ):
  global simu, comp, robot, grasp_container, grasp_association
  part = DetectionHelper( tool_name, bound )
  if part:
    iface = comp.findBehaviour(iface_name)
    if iface:
      # check interfaces in the other component
      ifaces = part.findBehavioursByType(VC_ONETOONEINTERFACE )
      for other_iface in ifaces:
        if iface.canConnect(other_iface ):
          if iface.connect( other_iface, True ):
            return "Ok"
          else:
            return "Tool to mount has interface already connected"
      return "No matching interface detected to mount tool"
  else:
    return "No tool detected to mount"


def ReleaseHelper( part, detection_tolerance ):
  global simu, comp, robot
  part.update()
  # Detect where to release using offset at "gravity" direction
  # Basically just translate detected component in the world z axis direction downwards with tolerance value
  gravity = vcVector.new(0,0,-detection_tolerance)
  # Calculate gravity at node parent coordinate system ( reset translate component and use just rotation)
  invert_coord = part.Parent.InverseWorldPositionMatrix
  invert_coord.P = vcVector.new()
  node_gravity = invert_coord * gravity
  # Translate node using gravity
  # rint "Node gravity X=%f Y=%f Z=%f" % (node_gravity.X,node_gravity.Y,node_gravity.Z)
  orig_m = part.PositionMatrix
  pos_m = part.PositionMatrix
  pos = pos_m.P
  pos += node_gravity
  pos_m.P = pos
  part.PositionMatrix = pos_m
  if UpdateWorldIOnRelease.Value:
    simu.update()
  else:
    part.update()
  release_node = None
  detector = simu.newCollisionDetector()
  detector.NodeListA = [ (part, VC_NODELIST_INCLUDE, VC_NODELIST_COMPONENT ) ]
  detector.NodeListB = [ (simu.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) , (comp, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE )]
  hit = detector.testOneCollision()
  # Restore location (minus gravity offset)
  part.PositionMatrix = orig_m
  part.update()
  if hit:
    return detector.getHitNodeB(0)
  else:
    return None


def Release( output, detection_tolerance ):
  global simu, comp, robot, grasp_container, grasp_association, grasped_components
  # Find first the part to be released
  physicsType = None
  set_phys = False
  part = None
  if len(grasped_components) == 0:
    return "No component to release"
  release_list = []
  if ASSOCIATE_GRASPING_TO_OUTPUT:
    try:
      if MultiGrasp.Value:
        list = grasp_association[output]
        if list:
          release_list = list
      else:
        part = grasp_association[output]
        release_list.append(part)
    except:
      pass
    if len(release_list)==0:
      return "No component associated to release with output %i" % output
  else:
    if MultiGrasp.Value:
      release_list = grasped_components
    else:
      release_list.append(grasped_components[0])
  grasp_association[output] = None
  count_to_physics = 0
  count_to_node = 0
  count_to_world = 0
  count_to_container = 0
  count_to_world_container = 0
  for part in release_list:
    grasped_components.remove( part )
    if ReleaseToPhysics.Value == 'Always':
      set_phys = True
      release_node = simu.World
    elif ReleaseToWorld.Value:
      release_node = simu.World
      if ReleaseToPhysics.Value != 'Never':
        set_phys = True
    else:
      release_node = ReleaseHelper( part, detection_tolerance )
      if not release_node:
        if ReleaseToPhysics.Value != 'Never':
          set_phys = True
        release_node = simu.World
    # Calculate position in relative to the detected component (NOTE: component origin)
    pos_m = release_node.InverseWorldPositionMatrix * part.WorldPositionMatrix
    pos = pos_m.P
    # Extract all containers what this node owns (containers where to release)
    ContainerList = release_node.getBehavioursByType(VC_CONTAINER)
    # Search closest container in the detected component
    BestContainer = None
    BestProximity = 1e10
    for cont in ContainerList:
      dist = cont.getContainerProximity( pos )
      if dist < 0.0: # Container with no physical limits return -1.0 value
        dist = 1e5
      if dist < BestProximity:
        BestContainer = cont
        BestProximity = dist
    # Release component to the closest container
    if BestContainer:
      # rint "Container \'%s\' found at proximity = %f" % (BestContainer.Name, BestProximity)
      BestContainer.grab( part )
      str = ""
      if release_node.Parent:
        count_to_container += 1
        str = "Ok, released to node container"
      else:
        count_to_world_container += 1
        str = "Ok, released to world container"
      if not MultiGrasp.Value:
        if set_phys:
          str = "Ok, released to physics"
          setPhysics(part)
    else:
      # rint "No container found to release component, just connect it hierarchically"
      pos_m = release_node.InverseWorldPositionMatrix * part.WorldPositionMatrix
      part.PositionMatrix = pos_m
      release_node.attach(part, False)
      str = ""
      if release_node.Parent:
        count_to_node += 1
        str = "Ok, attached directly to node"
      else:
        count_to_world += 1
        str = "Ok, attached directly to world"
      if not MultiGrasp.Value:
        if set_phys:
          str = "Ok, released to physics"
          setPhysics(part)
        return str
    if set_phys:
      count_to_physics += 1
      setPhysics(part)
  str = "Ok, "
  if count_to_physics:
    str += "released %i to physics. " % count_to_physics
  else:
    if count_to_node == 1:
      str += "attached directly to node, "
    elif count_to_node > 1:
          str += "attached %i directly to node, " % count_to_node
    if count_to_world == 1:
      str += "attached directly to world, "
    elif count_to_world > 1:
          str += "attached %i directly to world, " % count_to_world
    if count_to_container == 1:
      str += "released to node container, "
    elif count_to_container > 1:
          str += "released %i to node container, " % count_to_container
    if count_to_world_container == 1:
      str += "released to world container, "
    elif count_to_world_container > 1:
          str += "released %i to world container, " % count_to_world_container
  return str


def setPhysics(part):
  root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
  if root_phys_entities:
    root_phys_entity = root_phys_entities[0]
  else:
    root_phys_entity = part.createBehaviour('rPhysicsEntity', 'PhysicsEntity')
  physicsType = root_phys_entity.getProperty('Physics Type')
  physicsType.Value = 0


def UnmountTool( output, detection_tolerance, iface_name ):
  global simu, comp, robot, grasp_container
  # Look first the gripper to be released
  iface = comp.findBehaviour(iface_name)
  if not iface:
    return
  gripper = iface.ConnectedComponent
  if not gripper:
    return "No tool to unmount"
  # Unconnect interface
  if UpdateWorldIOnRelease.Value:
    simu.update()
  else:
    gripper.update()
  world_pos = gripper.WorldPositionMatrix
  other_iface = iface.ConnectedSection.ConnectedToSection.Interface
  # look if gripper can be released to some other component pallet etc.
  release_node = ReleaseHelper( gripper, detection_tolerance )
  if not release_node:
    if not ReleaseToWorld.Value:
      iface.unConnect(other_iface)
      return "No component detected to unmount tool to"
    release_node = simu.World
  # Calculate position in relative to the detected component (NOTE: component origin)
  pos_m = release_node.InverseWorldPositionMatrix * gripper.WorldPositionMatrix
  pos = pos_m.P
  iface.unConnect(other_iface)
  # Extract all containers what this node owns (containers where to release)
  ContainerList = release_node.getBehavioursByType(VC_CONTAINER)
  # Search closest container in the detected component
  BestContainer = None
  BestProximity = 1e10
  for cont in ContainerList:
    dist = cont.getContainerProximity( pos )
    if dist < 0.0: # Container with no physical limits return -1.0 value
      dist = 1e5
    if dist < BestProximity:
      BestContainer = cont
      BestProximity = dist
  # Release component to the closest container
  if BestContainer:
    # rint "Container \'%s\' found at proximity = %f" % (BestContainer.Name, BestProximity)
    BestContainer.grab( gripper )
    if release_node.Parent:
      return "Ok, unmounted to node container"
    else:
      return "Ok, unmounted to world container"
  else:
    # rint "No container found to release component"
    gripper.update()
    pos_m = release_node.InverseWorldPositionMatrix * world_pos
    gripper.PositionMatrix = pos_m
    release_node.attach(gripper, False)
    if release_node.Parent:
      return "Ok, unmounted directly to node"
    else:
      return "Ok, unmounted directly to world"


def TraceOn( port, tool_name, node_name, color ):
  global simu, comp, robot, grasp_container, traces, trace_lines, world
  # Find out if trace is already active
  for trace in traces:
    if trace.Port == port:
      return
  # Following code sets a specific tool and extracts tool matrix in world coordinate system
  # This transformation is then applied to the bound box defining volume.
  mat = app.findMaterial(color)
  layer = app.findLayer(TRACE_LAYER)
  world = simu.World
  container = None
  cnode = None
  if node_name != "":
    parts = node_name.split("::",2)
    if len(parts) == 1:
      ctrcmp = comp
      nname = parts[0]
    else:
      ctrcmp = app.findComponent(parts[0])
      nname = parts[1]
    if ctrcmp != None:
      cnode = ctrcmp.findNode(nname)
    else:
      cnode = None
    if cnode != None:
      container = cnode.UserGeometry
  if container == None:
    container = world.UserGeometry
  if cnode == None:
    cnode = world
  # Create trace object, which is used at updating
  trace = Trace()
  trace.Tool = Tool(tool_name)
  trace.Port = port
  trace.ToolName = tool_name
  trace.Base = Base(tool_name)
  trace.BaseName = tool_name
  trace.Node = cnode
  trace.Material = color
  trace.LineSet = container.createGeometrySet(VC_COMPACTLINESET)
  trace_lines.append(trace.LineSet)
  trace.LineSet.Name = "TRACE"
  trace.LineSet.Material = mat
  trace.LineSet.Layer = layer
  trace.LineSet.LineWidth = TraceWidth.Value
  trace.Line = []
  trace.Robot = robot
  '''
  # light effects
  if port < 32:
    trace.Particles = app.findComponent( "Arc F/X "+comp.Name )
    light = app.findLight( "Torch"+comp.Name )
    if not light:
      light = app.createLight("Torch"+comp.Name)
    light.Type = VC_LIGHT_POINT
    light.Enabled = True
    light.QuadraticAttenuation = 0.0001
    light.Intensity = 10.0
    trace.Light = light'''
  m = CalcTracePosition( trace )
  trace.Line.append(m.P)
  trace.LineSet.addLine(trace.Line)
  trace.LineSet.update()
  traces.append(trace)
  resumeRun()
  return "Ok"


def TraceOff( tool_name ):
  global simu, comp, robot, grasp_container, traces
  trace = None
  for t in traces:
    if t.ToolName == tool_name:
      trace = t
      break
    if t.BaseName == tool_name:
      trace = t
      break
  if trace:
    '''
    # light effects
    if trace.Light:
      trace.Light.Enabled = False'''
    m = CalcTracePosition( trace )
    trace.Node.update()
    trace.Line = trace.LineSet.getLine(trace.LineSet.LineCount - 1)
    if trace.Line:
      trace.LineSet.deleteLine(trace.LineSet.LineCount - 1)
    else:
      trace.Line = []
    trace.Line.append(m.P)
    trace.LineSet.addLine(trace.Line)
    trace.LineSet.update()
    traces.remove(trace)
  CheckIfSuspendRun()
  return "Ok"


def SweptVolumeOn( port, action):
  global activeSweptVolumeItems, allSweptVolumeItems
  ready = False
  svName = "%s_SweptVolume_%i" % (comp.Name, port)
  # FIND THE LAYOUT ITEM and create new if not found
  sv = app.findLayoutItem(svName)
  if(not sv):
    sv = app.createLayoutItem(VC_LAYOUTITEM_IT_SWEPTVOLUME)
    sv.Name = svName;
  # SET SWEPT VOLUME PROPERTIES
  sv.Material = action.Material
  sv.Decomposition = VC_CONVEXHULL  # # VC_DECOMPOSITION | VC_CONVEXHULL | VC_CONVEXHULL_TRIMESH
  if action.SweptMethod == "HullFromHulls":
    sv.method = VC_SWEPT_HULL_FROM_HULLS
  elif action.SweptMethod == "HullFromSilhouettes":
    sv.method = VC_SWEPT_HULL_FROM_SILHOUETTES
  else:
    sv.method = VC_SWEPT_LOFTING
  if not action.SweptGeometry:
    allSweptVolumeItems.append(sv)
    return "Please include node for Geometry"
  swept_nodes_list = action.SweptGeometry  
  if swept_nodes_list == [comp]:
    # add the whole kinematic chain of the robot (including eoat) if only the robot root node is configured
    kids = comp.Children
    while kids:
      nnode = kids.pop(0)
      kids.extend(nnode.Children)
      swept_nodes_list.append(nnode)
  sv.InputNodeList = [(x,VC_NODELIST_INCLUDE,VC_NODELIST_NODE) for x in swept_nodes_list]
  # DEFINE node where to attach the wsept volume
  if action.AttachTo == "Node":
    sv.TargetNode = action.TargetNode
  elif action.AttachTo == "Closest":
    tool_name = action.UsingTool
    bound = action.DetectionVolumeSize
    target_node = DetectionHelper( tool_name, bound )
    if not target_node:
      allSweptVolumeItems.append(sv)
      print 'Target node for attaching the swept volume not detected.'
      return
    else:
      sv.TargetNode = target_node
  else:
    #create new component
    c = app.findComponent(svName + "_result") 
    if not c: # new component is created only if existing component is not found
      c = app.createComponent()
      c.Name = svName + "_result"
    sv.TargetNode = c
  # if all is set, then add it
  activeSweptVolumeItems.append(sv)
  allSweptVolumeItems.append(sv)
  sv.begin()
  resumeRun()
  return "Ok"


def SweptVolumeOff(port, store):
  global activeSweptVolumeItems
  svName = "%s_SweptVolume_%i" % (comp.Name, port)
  sv = app.findLayoutItem(svName)
  # sv might exists but it should be in activeSweptVolumeItems as well
  if sv and sv in activeSweptVolumeItems:
    sv.end()
    activeSweptVolumeItems.remove(sv)   
    if store:
      sv.store() 
    #sv.clear()
  CheckIfSuspendRun()


def CalcTracePosition( trace ):
  if trace.Node != world:
    trace.Node.update()
  # If trace is drawn by the tool
  if trace.Tool:
    tnode = trace.Tool.Node
    tnode.update()
    m = trace.Tool.PositionMatrix
    m = tnode.WorldPositionMatrix * m
  #if trace is drawn by the base
  elif trace.Base:
    tnode = trace.Base.Node
    if tnode:
      tnode.update()
      tnode_mtx = tnode.WorldPositionMatrix
    else:
      comp.update()
      tnode_mtx = comp.WorldPositionMatrix * controller.WorldTransformMatrix
    m = trace.Base.PositionMatrix
    m = tnode_mtx * m
  else:
    tnode = robot.FlangeNode
    tnode.update()
    m = vcMatrix.new()
    m = tnode.WorldPositionMatrix * m  
  '''
  # light effects
  light = trace.Light
  if light and light.Enabled:
    light.Position = m.P
    light.Direction = m.A
    particles = trace.Particles
    if particles:
      ppm = particles.PositionMatrix
      ppm.P = m.P
      particles.PositionMatrix = ppm'''
  if TraceZOffset.Value:
    m.translateRel( 0., 0., -TraceZOffset.Value )
  return trace.Node.InverseWorldPositionMatrix * m


EPSILON = 10e-5


def CheckIfSuspendRun():
  global traces, activeSweptVolumeItems
  if (len(traces) == 0 and len(activeSweptVolumeItems) == 0):
    suspendRun()


def OnRun():
  global simu, comp, robot, grasp_container, traces, activeSweptVolumeItems
  cached = vcVector.new(999999,999999,999999)
  CheckIfSuspendRun()
  while True:
    for trace in traces:
      m = CalcTracePosition( trace )
      if (cached-m.P).length() > EPSILON:
        cached = m.P
        trace.Line = trace.LineSet.getLine(trace.LineSet.LineCount - 1)
        if trace.Line:
          trace.LineSet.deleteLine(trace.LineSet.LineCount - 1)
        else:
          trace.Line = []
        trace.Line.append(m.P)
        trace.LineSet.addLine(trace.Line)
        trace.LineSet.update()
    delay(min(0.05,simu.SimSpeed))
    for sv in activeSweptVolumeItems:
      sv.next()   
      delay(0.05) # for the moment is hard coded


def OnReset():
  global simu, comp, robot, grasp_container, trace, trace_lines, activeSweptVolumeItems, allSweptVolumeItems
  '''
  # light effects
  light = app.findLight( "Torch"+comp.Name )
  if light:
    light.Enabled = False'''
  if simu:
    for line in trace_lines:
      ctr = line.Container
      ctr.deleteGeometrySet(line)
    trace_lines = []
  for sv in activeSweptVolumeItems:
    sv.end()
    #sv.clear()
  for  sv in allSweptVolumeItems:
    sv.clear()
    app.deleteLayoutItem(sv)    
  activeSweptVolumeItems = []
  allSweptVolumeItems = []


UpdateWorldIOnRelease = comp.getProperty('SignalActions::UpdateWorldIOnRelease')
if not UpdateWorldIOnRelease:
  UpdateWorldIOnRelease = comp.createProperty(VC_BOOLEAN, 'SignalActions::UpdateWorldIOnRelease')

AutoConfigure()
